@@ -17,8 +17,12 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <sys/types.h>
+#include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
+#include <glib-object.h>
+#include <glib/gstdio.h>
 #include <gtk/gtk.h>
 #include "gailtextview.h"
 #include <libgail-util/gailmisc.h>
@@ -124,6 +128,16 @@ static void       gail_text_view_paste_text            (AtkEditableText  *text,
 static void       gail_text_view_paste_received        (GtkClipboard     *clipboard,
                                                         const gchar      *text,
                                                         gpointer         data);
+/* AtkStreamableContent */
+static void       atk_streamable_content_interface_init    (AtkStreamableContentIface *iface);
+static gint       gail_streamable_content_get_n_mime_types (AtkStreamableContent *streamable);
+static G_CONST_RETURN gchar* gail_streamable_content_get_mime_type (AtkStreamableContent *streamable,
+								    gint i);
+static GIOChannel* gail_streamable_content_get_stream       (AtkStreamableContent *streamable,
+							     const gchar *mime_type);
+/* getURI requires atk-1.12.0 or later
+static void       gail_streamable_content_get_uri          (AtkStreamableContent *streamable);
+*/
 
 /* Callbacks */
 
@@ -152,6 +166,7 @@ static gint             get_insert_offset              (GtkTextBuffer    *buffer
 static gint             get_selection_bound            (GtkTextBuffer    *buffer);
 static void             emit_text_caret_moved          (GailTextView     *gail_text_view,
                                                         gint             insert_offset);
+static gint             insert_handler                 (gpointer         data);
 static gint             insert_idle_handler            (gpointer         data);
 
 static GailWidgetClass *parent_class = NULL;
@@ -199,12 +214,21 @@ gail_text_view_get_type (void)
         NULL
       };
 
+      static const GInterfaceInfo atk_streamable_content_info =
+      {
+        (GInterfaceInitFunc) atk_streamable_content_interface_init,
+        (GInterfaceFinalizeFunc) NULL,
+        NULL
+      };
+
       type = g_type_register_static (GAIL_TYPE_CONTAINER,
                                      "GailTextView", &tinfo, 0);
       g_type_add_interface_static (type, ATK_TYPE_EDITABLE_TEXT,
                                    &atk_editable_text_info);
       g_type_add_interface_static (type, ATK_TYPE_TEXT,
                                    &atk_text_info);
+      g_type_add_interface_static (type, ATK_TYPE_STREAMABLE_CONTENT,
+                                   &atk_streamable_content_info);
     }
 
   return type;
@@ -1421,7 +1445,7 @@ _gail_text_view_insert_text_cb (GtkTextBuffer *buffer,
 
   gail_text_view->signal_name = "text_changed::insert";
   position = gtk_text_iter_get_offset (arg1);
-  length = arg3;
+  length = g_utf8_strlen(arg2, arg3);
   
   if (gail_text_view->length == 0)
     {
@@ -1442,7 +1466,7 @@ _gail_text_view_insert_text_cb (GtkTextBuffer *buffer,
           g_source_remove (gail_text_view->insert_notify_handler);
         }
       gail_text_view->insert_notify_handler = 0;
-      insert_idle_handler (gail_text_view);
+      insert_handler (gail_text_view);
       gail_text_view->position = position;
       gail_text_view->length = length;
     }
@@ -1487,7 +1511,7 @@ _gail_text_view_delete_range_cb (GtkTextBuffer *buffer,
           return;
         }
 
-      insert_idle_handler (gail_text_view);
+      insert_handler (gail_text_view);
     }
   g_signal_emit_by_name (accessible, "text_changed::delete",
                          offset, length);
@@ -1635,7 +1659,7 @@ emit_text_caret_moved (GailTextView *gail_text_view,
     {
       g_source_remove (gail_text_view->insert_notify_handler);
       gail_text_view->insert_notify_handler = 0;
-      insert_idle_handler (gail_text_view);
+      insert_handler (gail_text_view);
     }
 
   if (insert_offset != gail_text_view->previous_insert_offset)
@@ -1652,13 +1676,11 @@ emit_text_caret_moved (GailTextView *gail_text_view,
 }
 
 static gint
-insert_idle_handler (gpointer data)
+insert_handler (gpointer data)
 {
   GailTextView *gail_text_view;
   GtkTextBuffer *buffer;
 
-  GDK_THREADS_ENTER ();
-
   gail_text_view = GAIL_TEXT_VIEW (data);
 
   g_signal_emit_by_name (data,
@@ -1680,7 +1702,125 @@ insert_idle_handler (gpointer data)
       gail_text_view->previous_selection_bound = get_selection_bound (buffer);
     }
 
+  return FALSE;
+}
+
+static gint
+insert_idle_handler (gpointer data)
+{
+  GDK_THREADS_ENTER ();
+  insert_handler (data);
   GDK_THREADS_LEAVE ();
 
   return FALSE;
 }
+
+static void       
+atk_streamable_content_interface_init    (AtkStreamableContentIface *iface)
+{
+  g_return_if_fail (iface != NULL);
+  
+  iface->get_n_mime_types = gail_streamable_content_get_n_mime_types;
+  iface->get_mime_type = gail_streamable_content_get_mime_type;
+  iface->get_stream = gail_streamable_content_get_stream;
+}
+
+static gint       gail_streamable_content_get_n_mime_types (AtkStreamableContent *streamable)
+{
+    gint n_mime_types = 0;
+
+    if (GAIL_IS_TEXT_VIEW (streamable) && GAIL_TEXT_VIEW (streamable)->textutil)
+    {
+	int i;
+	gboolean advertises_plaintext = FALSE;
+	GdkAtom *atoms =
+	     gtk_text_buffer_get_serialize_formats (
+		GAIL_TEXT_VIEW (streamable)->textutil->buffer, 
+		&n_mime_types);
+	for (i = 0; i < n_mime_types-1; ++i)
+	    if (!strcmp ("text/plain", gdk_atom_name (atoms[i])))
+		advertises_plaintext = TRUE;
+	if (!advertises_plaintext) ++n_mime_types; 
+        /* we support text/plain even if the GtkTextBuffer doesn't */
+    }
+    return n_mime_types;
+}
+
+static G_CONST_RETURN gchar*       
+gail_streamable_content_get_mime_type (AtkStreamableContent *streamable, gint i)
+{
+    if (GAIL_IS_TEXT_VIEW (streamable) && GAIL_TEXT_VIEW (streamable)->textutil)
+    {
+	gint n_mime_types = 0;
+	GdkAtom *atoms;
+	atoms = gtk_text_buffer_get_serialize_formats (
+	    GAIL_TEXT_VIEW (streamable)->textutil->buffer, 
+	    &n_mime_types);
+	if (i < n_mime_types)
+	{
+	    return gdk_atom_name (atoms [i]);
+	}
+	else if (i == n_mime_types)
+	    return "text/plain";
+    }
+    return NULL;
+}
+
+static GIOChannel*       gail_streamable_content_get_stream       (AtkStreamableContent *streamable,
+								   const gchar *mime_type)
+{
+    gint i, n_mime_types = 0;
+    GdkAtom *atoms;
+    if (!GAIL_IS_TEXT_VIEW (streamable) || !GAIL_TEXT_VIEW (streamable)->textutil)
+	return NULL;
+    atoms = gtk_text_buffer_get_serialize_formats (
+	GAIL_TEXT_VIEW (streamable)->textutil->buffer, 
+	&n_mime_types);
+    for (i = 0; i < n_mime_types; ++i) 
+    {
+	if (!strcmp ("text/plain", mime_type) ||
+	    !strcmp (gdk_atom_name (atoms[i]), mime_type)) {
+	    GtkTextBuffer *buffer;
+	    guint8 *cbuf;
+	    GError *err = NULL;
+	    gsize len, written;
+	    gchar tname[80];
+	    GtkTextIter start, end;
+	    GIOChannel *gio = NULL;
+	    int fd;
+	    buffer = GAIL_TEXT_VIEW (streamable)->textutil->buffer;
+	    gtk_text_buffer_get_iter_at_offset (buffer, &start, 0);
+	    gtk_text_buffer_get_iter_at_offset (buffer, &end, -1);
+	    if (!strcmp ("text/plain", mime_type)) 
+	    {
+		cbuf = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
+		len = strlen (cbuf); 
+	    }
+	    else
+	    {
+		cbuf = gtk_text_buffer_serialize (buffer, buffer, atoms[i], &start, &end, &len);
+	    }
+	    g_snprintf (tname, 20, "streamXXXXXX");
+	    fd = g_mkstemp (tname);
+	    gio = g_io_channel_unix_new (fd);
+	    g_io_channel_set_encoding (gio, NULL, &err);
+	    if (!err) g_io_channel_write_chars (gio, cbuf, (gssize) len, &written, &err);
+	    else g_message (err->message);
+	    if (!err) g_io_channel_seek_position (gio, 0, G_SEEK_SET, &err);
+	    else g_message (err->message);
+	    if (!err) g_io_channel_flush (gio, &err);
+	    else g_message (err->message);
+	    if (err) {
+		g_message ("<error writing to stream [%s]>", tname);
+		g_free (err);
+	    }
+	    /* make sure the file is removed on unref of the giochannel */
+	    else {
+		g_unlink (tname);
+		return gio; 
+	    }
+	}
+    }
+    return NULL;
+}
+
